from __future__ import annotations

import inspect
import pathlib
from typing import Iterator

from .utils import is_path, split_line, split_lines


class Lines:

    def __init__(
        self,
        template: str | pathlib.Path | None = None,
        *,
        parent: Line | None = None,
        stack_level: int = 0,
    ) -> None:
        self.parent = parent
        if not self.parent:
            self._path = None
            self._source_path, self._source_line_number = self._get_source(stack_level + 1)
        self.lines: list[Line] = []
        if template:
            if is_path(template):
                self._path = pathlib.Path(template)
                text = self._path.read_text()
                offset = 0
            else:
                text = str(template)
                offset = 1
            stack: list[Line] = []
            for number, line_text in split_lines(text):
                indent, content = split_line(line_text)
                line = Line(number - offset, indent, content)
                while stack and stack[-1].indent >= indent:
                    stack.pop()
                if stack:
                    stack[-1].children.append(line)
                else:
                    self.append(line)
                stack.append(line)

    def __str__(self) -> str:
        if self.path:
            return f"{self.path.name} at {self.source}"
        return self.source

    def __repr__(self) -> str:
        return f"<{self}>"

    def __bool__(self) -> bool:
        return bool(self.lines)

    def __len__(self) -> int:
        return len(self.lines)

    def __iter__(self) -> Iterator[Line]:
        yield from self.lines

    def __getitem__(self, index: int) -> Line:
        return self.lines[index]

    @property
    def path(self) -> pathlib.Path | None:
        if self.parent:
            return self.parent.path
        return self._path

    @property
    def source_path(self) -> pathlib.Path:
        if self.parent:
            return self.parent.source_path
        if self._source_path:
            return self._source_path
        raise RuntimeError("source path not set")

    @property
    def source_line_number(self) -> int:
        if self.parent and self.parent.container:
            return self.parent.container.source_line_number
        if self._source_line_number:
            return self._source_line_number
        raise RuntimeError("source line not set")

    @property
    def source(self) -> str:
        return f"{self.source_path.name}:{self.source_line_number}"

    def append(self, line: Line) -> None:
        self.lines.append(line)
        line.container = self

    def set_source(self, path: pathlib.Path, line_number: int) -> None:
        self._source_path, self._source_line_number = path, line_number

    def snap(self, to: int | None = None) -> Lines:
        if to is None:
            if self.parent:
                to = self.parent.indent
            else:
                to = 0
        for line in self.lines:
            line.dedent(line.indent - to)
        return self

    def to_string(self) -> str:
        output: list[str] = []
        for line in self.lines:
            output.append(line.to_string())
            if line.children:
                output.append(line.children.to_string())
        return "\n".join(output)

    def _get_source(self, stack_level: int) -> tuple[pathlib.Path, int]:
        frame = inspect.currentframe()
        for _ in range(stack_level + 1):
            frame = frame and frame.f_back
        if not frame:
            raise RuntimeError("unable to infer source")
        path = pathlib.Path(frame.f_code.co_filename)
        return path, frame.f_lineno


class Line:

    def __init__(self, number: int, indent: int, content: str) -> None:
        self.number = number
        self.indent = indent
        self.content = content
        self.container: Lines | None = None
        self.children = Lines(parent=self)

    def __str__(self) -> str:
        if self.path:
            return f"line {self.number} of {self.path.name} at {self.source}"
        return f"line {self.number} at {self.source}"

    def __repr__(self) -> str:
        return f"<{self}: {self.indent} | {self.content}>"

    @property
    def path(self) -> pathlib.Path | None:
        return self.container.path if self.container is not None else None

    @property
    def source_path(self) -> pathlib.Path:
        if not self.container:
            raise RuntimeError("source path not set")
        return self.container.source_path

    @property
    def source_line_number(self) -> int:
        if not self.container:
            raise RuntimeError("source line not set")
        if self.container.path:
            return self.container.source_line_number
        return self.container.source_line_number + 1 + self.number

    @property
    def source(self) -> str:
        return f"{self.source_path.name}:{self.source_line_number}"

    def dedent(self, offset: int) -> Line:
        self.indent = max(self.indent - offset, 0)
        for line in self.children.lines:
            line.dedent(offset)
        return self

    def to_string(self) -> str:
        return " " * self.indent + self.content



from __future__ import annotations

import contextlib
import pathlib
import re
import sys
from typing import Any, Callable, ClassVar, Iterable, Iterator

from .collect import collect_definitions, collect_global_references
from .interpolate import interpolate as interpolate_
from .interpolate import parse_arguments
from .lines import Line, Lines
from .utils import and_, is_path, split_lines

type Transpiler = Callable[[Junk, str], None]
type MetaCallback = Callable[[Junk], None]
type MetaModule = str | pathlib.Path | dict[str, Any] | Iterable[MetaModule]


META_REGEX = re.compile(
    r"""
    ^
    ([a-zA-Z_][a-zA-Z0-9_]*)
    (?:
        (:{0,2})
        \s+
        (.*)
    )?
    $
    """,
    flags=re.VERBOSE,
)


class Junk:

    code_prefix: ClassVar[str] = "!"
    meta_prefix: ClassVar[str] = "%"
    comment_prefix: ClassVar[str] = "#"
    meta_function_prefix: ClassVar[str] = "meta_"
    eval_function_prefix: ClassVar[str] = "eval_"
    on_load_function_name: ClassVar[str] = "on_load"
    builtins_directories: ClassVar[list[pathlib.Path]] = [pathlib.Path(__file__).parent / "builtins"]
    common_module_name: ClassVar[str] = "common"

    add_source_comments_by_default: ClassVar[bool] = True
    load_common_by_default: ClassVar[bool] = True
    transpile_standalone_by_default: ClassVar[bool] = False
    default_interpolation: ClassVar[str] = "{ }"
    interpolate_by_default: ClassVar[bool] = True

    EMIT: ClassVar[str] = "emit"
    INDENT: ClassVar[str] = "indent"
    STOP_EVALUATION: ClassVar[str] = "StopEvaluation"

    def __init__(
        self,
        template: str | pathlib.Path | None = None,
        *,
        load_common: bool | None = None,
        add_source_comments: bool | None = None,
        stack_level: int = 0,
    ) -> None:
        if load_common is None:
            load_common = self.load_common_by_default
        if add_source_comments is None:
            add_source_comments = self.add_source_comments_by_default
        self.lines = Lines(template, stack_level=stack_level + 1)
        self.path = self.lines.path or self.lines.source_path
        self.add_source_comments = add_source_comments
        self.code_indent = 0
        self.text_indent = 0
        self.eval_indent = 0
        self.meta_lines: list[Any] = []
        self.eval_lines: list[Any] = []
        self.transpilers: dict[str, Transpiler] = {
            self.code_prefix: code,
            self.meta_prefix: meta,
            "": text,
        }
        self.meta_context: dict[str, Any] = {}
        self.meta_namespace: dict[str, Any] = {
            "junk": self,
            "load": load,
        }
        self.meta_state: dict[str, Any] = {}
        self.meta_callbacks: list[MetaCallback] = []
        self.eval_context: dict[str, Any] = {}
        self.eval_namespace: dict[str, Any] = {
            "junk": self,
            self.EMIT: self.emit,
            self.INDENT: self.indent,
            self.STOP_EVALUATION: StopEvaluation,
        }
        self.eval_state: dict[str, Any] = {}
        self.interpolation: str = self.default_interpolation
        self.inline: bool = False
        self._active_lines: list[Line] = []
        if load_common:
            self.load(self.common_module_name)

    def __str__(self) -> str:
        if self.path != self.lines.source_path:
            return f"junk of {self.path} at {self.source}"
        return f"junk at {self.source}"

    def __repr__(self) -> str:
        return f"<{self}>"

    @property
    def source_path(self) -> pathlib.Path:
        return self.lines.source_path

    @property
    def source_line_number(self) -> int:
        return self.lines.source_line_number

    @property
    def source(self) -> str:
        return self.lines.source

    @property
    def has_line(self) -> bool:
        return bool(self._active_lines)

    @property
    def line(self) -> Line:
        return self._active_lines[-1]

    def load(self, target: MetaModule) -> None:
        load(self, target)

    def transpile(self, context: dict[str, Any] | None = None, /, **context_kwargs: Any) -> str:
        if context:
            self.meta_context.update(context)
        if context_kwargs:
            self.meta_context.update(context_kwargs)
        self.meta_namespace.update(self.meta_context)
        self.proceed(self.lines)
        for callback in self.meta_callbacks:
            callback(self)
        return self.to_string()

    def to_string(self, standalone: bool | None = None) -> str:
        if standalone is None:
            standalone = self.transpile_standalone_by_default
        meta = "\n".join(map(str, self.meta_lines))
        if standalone:
            meta = self._generate_intro(meta)
        return meta

    def evaluate(self, context: dict[str, Any] | None = None, /, **context_kwargs: Any) -> str:
        if context:
            self.eval_context.update(context)
        if context_kwargs:
            self.eval_context.update(context_kwargs)
        self.eval_namespace.update(self.eval_context)
        meta = self.to_string()
        try:
            exec(compile(meta, self.source, "exec"), self.eval_namespace)
        except StopEvaluation:
            pass
        except Exception as error:
            raise EvaluationError(self, error) from None
        return "".join(map(str, self.eval_lines)).rstrip()

    def proceed(self, lines: Lines | None = None) -> None:
        if lines is None:
            lines = self.line.children
        for line in lines:
            with self._set_active_line(line):
                for prefix, transpile in sorted(
                    self.transpilers.items(),
                    key=lambda x: len(x[0]),
                    reverse=True,
                ):
                    if line.content.startswith(prefix):
                        content = line.content.removeprefix(prefix).lstrip()
                        transpile(self, content)
                        break
                else:
                    transpilers = (f"{transpile.__name__} ({prefix})" for prefix, transpile in self.transpilers.items())
                    raise ValueError(f"unable to transpile {line} (considered {and_(transpilers)})")

    @contextlib.contextmanager
    def patch(self, **state: Any) -> Iterator[None]:
        prev_state: dict[str, Any] = {}
        for key, value in state.items():
            prev_state[key] = getattr(self, key)
            setattr(self, key, value)
        try:
            yield
        finally:
            for key, value in prev_state.items():
                setattr(self, key, value)

    def emit_code(self, code: str, add_source_comment: bool | None = None) -> None:
        if add_source_comment is None:
            add_source_comment = self.add_source_comments
        for _, code_line in split_lines(code):
            code_line = self.code_indent * " " + code_line
            if add_source_comment:
                code_line += self._source_comment
            self.meta_lines.append(code_line)

    @contextlib.contextmanager
    def increase_code_indent(self) -> Iterator[None]:
        with self.patch(code_indent=self.code_indent + 4):
            yield

    def emit_text(
        self,
        indent: int | None,
        text: str,
        interpolate: bool | None = None,
        newline: bool = True,
    ) -> None:
        if indent is not None:
            indent += self.text_indent
        if interpolate is None:
            interpolate = self.interpolate_by_default
        if not interpolate:
            args = [repr(text)]
        else:
            args = []
            for snippet, is_code in interpolate_(text, self.interpolation):
                if is_code:
                    args.append(f"{snippet}")
                else:
                    args.append(repr(snippet))
        if not newline:
            args.append("newline=False")
        if self.inline:
            args.append("inline=True")
        self.emit_code(f'{self.EMIT}({indent}, {", ".join(args)})')

    def emit_text_block(self, indent: int, text: str, interpolate: bool | None = None) -> None:
        if interpolate is None:
            interpolate = self.interpolate_by_default
        for _, line in split_lines(text):
            self.emit_text(indent, line, interpolate=interpolate)

    def interpolate(self, string: str) -> str:
        args = []
        for snippet, is_code in interpolate_(string, self.interpolation):
            if is_code:
                args.append(f"{snippet}")
            else:
                args.append(repr(snippet))
        if len(args) == 1:
            return f"str({args[0]})"
        return f'concat({", ".join(args)})'

    def derive(self, template: str | pathlib.Path, with_namespace: bool = False) -> Junk:
        if is_path(template, self.path.parent):
            template = self.path.parent / template
        junk = type(self)(template)
        if self.has_line:
            junk.lines.set_source(self.line.source_path, self.line.source_line_number)
        else:
            junk.lines.set_source(self.lines.source_path, self.lines.source_line_number)
        junk.add_source_comments = self.add_source_comments
        junk.meta_state = self.meta_state
        if with_namespace:
            junk.meta_namespace = self.meta_namespace.copy()
            junk.meta_namespace["junk"] = junk
        return junk

    def emit(
        self,
        indent: int | None,
        *args: Any,
        inline: bool = False,
        newline: bool = True,
    ) -> None:
        text = "".join(map(str, args))
        if inline:
            self.eval_lines.append(text)
        else:
            end = "\n" if newline else ""
            if indent is None:
                indent = 0
            else:
                indent += self.eval_indent
            self.eval_lines.append(f'{" " * indent}{text}{end}')

    @contextlib.contextmanager
    def indent(self, indent: int) -> Iterator[None]:
        self.eval_indent += indent
        try:
            yield
        finally:
            self.eval_indent -= indent

    @contextlib.contextmanager
    def _set_active_line(self, line: Line) -> Iterator[None]:
        self._active_lines.append(line)
        try:
            yield
        finally:
            self._active_lines.pop()

    @property
    def _source_comment(self) -> str:
        if self.line.path:
            return f"  # {self.line.path}:{self.line.number}"
        return f"  # {self.line.source_path}:{self.line.source_line_number}"

    def _generate_intro(self, meta: str) -> str:
        paths: set[pathlib.Path] = set()
        for name in collect_global_references(meta):
            if name not in self.eval_namespace:
                continue
            func = self.eval_namespace[name]
            if func is self or getattr(self, f"{self.eval_function_prefix}{name}", None) == func:
                continue
            while hasattr(func, "__wrapped__"):
                func = func.__wrapped__
            paths.add(pathlib.Path(func.__code__.co_filename))
        defs, imps = collect_definitions(meta, paths)
        intro: list[str] = []
        for name, (what, whence) in imps.items():
            if whence is None:
                if name == what:
                    intro.append(f"import {name}")
                else:
                    intro.append(f"import {what} as {name}")
            elif name == what:
                intro.append(f"from {whence} import {name}")
            else:
                intro.append(f"from {whence} import {what} as {name}")
        if intro:
            intro.append("")
        for name, def_ in defs.items():
            intro.append(f"{def_}\n")
        if intro:
            intro.append("")
        return "\n".join(intro) + meta


def code(junk: Junk, content: str) -> None:
    # comment line or block
    if content.startswith(junk.comment_prefix):
        return
    # code block
    if not content:
        code = junk.line.children.to_string()
        junk.emit_code(code)
        return
    # code line
    eval_indent = junk.line.indent
    if eval_indent:
        junk.emit_code(f"with {junk.INDENT}({eval_indent}):")
        junk.code_indent += 4
    junk.emit_code(content)
    with junk.increase_code_indent():
        junk.line.children.snap(0)
        junk.proceed()
    if eval_indent > 0:
        junk.code_indent -= 4


def meta(junk: Junk, content: str) -> None:
    # empty line
    if not content:
        junk.emit_text(0, "")
        return
    # meta code
    if content.startswith(junk.code_prefix):
        meta_code = content.removeprefix(junk.code_prefix).lstrip()
        # meta code block
        if not meta_code:
            meta_code = junk.line.children.snap(0).to_string()
            exec(meta_code, junk.meta_namespace)
            return
        # meta code line
        if not junk.line.children:
            exec(meta_code, junk.meta_namespace)
            return
        # meta code line with children
        meta_code += "\n    _()"
        exec(meta_code, {"_": lambda: junk.proceed(junk.line.children.snap())}, junk.meta_namespace)
        return
    # meta function
    match = META_REGEX.match(content)
    if not match:
        raise ValueError(
            f"expected meta function on {junk.line} to be '<function> [argument]', '<function>: <arguments>' or "
            f"'<function>:: <invocation>', but got {content!r}"
        )
    name, call_type, arg = match.groups()
    meta_functions = [name for name, value in junk.meta_namespace.items() if callable(value)]
    if name not in meta_functions:
        raise ValueError(
            f"unknown meta function {name!r} on {junk.line} " f"(available meta functions are {and_(meta_functions)})"
        )
    if not arg:
        meta_code = f"{name}(junk)"
    elif not call_type:
        meta_code = f"{name}(junk, {repr(arg)})"
    elif call_type == ":":
        args = ", ".join(parse_arguments(arg))
        meta_code = f"{name}(junk, {args})"
    else:
        meta_code = f"{name}(junk, {arg})"
    eval(meta_code, junk.meta_namespace)


def text(junk: Junk, content: str) -> None:
    if content:
        junk.emit_text(junk.line.indent, content)
    junk.proceed()


def load(junk: Junk, target: MetaModule) -> None:
    if isinstance(target, dict):
        namespace = target
    elif isinstance(target, str | pathlib.Path):
        path = junk.path.parent / target
        if not path.exists():
            for builtin_directory in junk.builtins_directories:
                builtin_path = builtin_directory / f"{target}.py"
                if builtin_path.exists():
                    path = builtin_path
                    break
            else:
                builtin_modules = []
                for builtin_directory in junk.builtins_directories:
                    for builtin_module in builtin_directory.glob("*.py"):
                        builtin_modules.append(builtin_module.stem)
                raise ValueError(
                    f"could not load {target!r} "
                    f"({path} does not exist, and available builtins are {and_(sorted(builtin_modules))})"
                )
        sys_path = sys.path.copy()
        sys.path.append(str(path.parent))
        try:
            text = path.read_text()
            code = compile(text, str(path), "exec")
            namespace = {}
            exec(code, namespace)
        finally:
            sys.path = sys_path
    else:
        for meta_module in target:
            load(junk, meta_module)
        return
    for key, value in namespace.items():
        if key.startswith(junk.meta_function_prefix):
            name = key.removeprefix(junk.meta_function_prefix)
            junk.meta_namespace[name] = value
        if key.startswith(junk.eval_function_prefix):
            name = key.removeprefix(junk.eval_function_prefix)
            junk.eval_namespace[name] = value.__get__(junk)
    if junk.on_load_function_name in namespace:
        namespace[junk.on_load_function_name](junk)


from .errors import EvaluationError, StopEvaluation



from .api import evaluate, render, transpile
from .junk import EvaluationError, Junk
from .lines import Line, Lines

__all__ = [
    "evaluate",
    "EvaluationError",
    "Junk",
    "Lines",
    "Line",
    "render",
    "transpile",
]



import pathlib
import re
from typing import Any, Iterable, Iterator, TypeGuard

LEADING_EMPTY_LINES = re.compile(r"^([ \t]*\r?\n)+")
LINE_REGEX = re.compile(r"^(\s*)(.*)$")
OPEN_LINE_SUFFIX = "\\"


def concat(iterable: Iterable[Any], conjunction: str, quote: bool = False) -> str:
    items = list(iterable)
    if quote:
        cast = repr
    else:
        cast = str
    if not items:
        return "<none>"
    if len(items) == 1:
        return cast(items[0])
    if len(items) == 2:
        return f"{cast(items[0])} {conjunction} {cast(items[1])}"
    return ", ".join(map(cast, items[:-1])) + f" {conjunction} {cast(items[-1])}"


def and_(iterable: Iterable[Any], quote: bool = False) -> str:
    return concat(iterable, "and", quote)


def or_(iterable: Iterable[Any], quote: bool = False) -> str:
    return concat(iterable, "or", quote)


def split_line(text: str) -> tuple[int, str]:
    whitespace, content = LINE_REGEX.match(text).groups()  # type: ignore
    indent = len(whitespace)
    return indent, content


def split_lines(text: str) -> Iterator[tuple[int, str]]:
    match = LEADING_EMPTY_LINES.match(text)
    if not match:
        skipped_lines = 0
    else:
        skipped_lines = match.group().count("\n")
        text = text[match.end() :]
    text = text.rstrip().expandtabs()
    indent: int | None = None
    open_line: list[str] = []
    open_line_number = 0
    for number, line in enumerate(text.splitlines(), skipped_lines + 1):
        if indent is None:
            indent, content = split_line(line)
            if content.endswith(OPEN_LINE_SUFFIX):
                open_line.append(content.removesuffix(OPEN_LINE_SUFFIX).strip())
                open_line_number = number
                continue
            yield number, content
            continue
        if open_line:
            if not line.strip():
                continue
            open_line.append(line.removesuffix(OPEN_LINE_SUFFIX).strip())
            if line.endswith(OPEN_LINE_SUFFIX):
                continue
            yield open_line_number, " ".join(open_line)
            open_line.clear()
            continue
        if not line.strip():
            yield number, ""
            continue
        prefix = line[:indent]
        if prefix and not prefix.isspace():
            raise ValueError(f"expected line {number} to start with {indent!r} spaces, but got {prefix!r}")
        line = line[indent:]
        if line.endswith(OPEN_LINE_SUFFIX):
            open_line.append(line.removesuffix(OPEN_LINE_SUFFIX).strip())
            open_line_number = number
            continue
        yield number, line
    if open_line:
        yield open_line_number, " ".join(open_line)


def is_path(path: str | pathlib.Path, parent: pathlib.Path | None = None) -> TypeGuard[pathlib.Path]:
    if isinstance(path, pathlib.Path):
        return True
    if "\n" in path:
        return False
    if parent:
        return (parent / path).exists()
    return pathlib.Path(path).exists()



from __future__ import annotations

import ast
import pathlib
import re

SOURCEMAP_REGEX = re.compile(r"(.*?)  # (.*?):(\d+)$")


class StopEvaluation(Exception):
    pass


class EvaluationError(Exception):

    def __init__(self, junk: Junk, error: Exception) -> None:
        self.junk = junk
        self.error = error

    def __str__(self) -> str:
        output: list[str] = [f"Failed to evaluate junk at {self.junk.source}."]
        output.append("Context:")
        if self.junk.eval_context:
            for key, value in self.junk.eval_context.items():
                output.append(f"  {key}: {value!r}")
        else:
            output.append("  <none>")
        output.append("Traceback (most recent call last):")
        traceback = self.error.__traceback__ and self.error.__traceback__.tb_next
        while traceback:
            try:
                line, template = self._parse_source(traceback.tb_frame.f_code.co_filename, traceback.tb_lineno)
            except Exception as error:
                line, template = f"? ({error})", None
            if traceback.tb_frame.f_code.co_filename == self.junk.source:
                file = "Junk"
            else:
                file = f'File "{traceback.tb_frame.f_code.co_filename}"'
            output.append(
                self._indent(2, f"{file}, line {traceback.tb_lineno}, in {traceback.tb_frame.f_code.co_name}")
            )
            output.append(self._indent(4, line))
            if template:
                template_line, template_path, template_line_number = template
                output.append(self._indent(4, f'@ File "{template_path}", line {template_line_number}'))
                output.append(self._indent(8, template_line))
            traceback = traceback.tb_next
        output.append(f"{type(self.error).__name__}: {self.error}")
        return "\n".join(output)

    def _parse_source(self, filename: str, line_number: int) -> tuple[str, tuple[str, pathlib.Path, int] | None]:
        if filename == self.junk.source:
            source = self.junk.to_string()
        else:
            path = pathlib.Path(filename)
            if not path.exists():
                raise ValueError(f"file {path} does not exist")
            source = path.read_text()
        lines = source.splitlines()
        if line_number > len(lines):
            raise ValueError(f"file {path} has only {len(lines)} lines, unable to find line {line_number}")
        line = lines[line_number - 1].strip()
        match = SOURCEMAP_REGEX.match(line)
        if match:
            line, template_path, template_line_number = match.groups()
            template_path = pathlib.Path(template_path)
            template_line_number = int(template_line_number)
            if not template_path.exists():
                raise ValueError(f"template {template_path} does not exist")
            template_lines = template_path.read_text().splitlines()
            if template_line_number > len(template_lines):
                raise ValueError(
                    f"template {template_path} has only {len(template_lines)} lines, "
                    f"unable to find line {template_line_number}"
                )
            template_line = template_lines[template_line_number - 1].strip()
            template = template_line, template_path, template_line_number
        else:
            template = None
        tree = ast.parse(source)
        for node in ast.walk(tree):
            if isinstance(node, ast.FunctionDef):
                # end_lineno can be None in older Python versions or for certain AST nodes
                if node.end_lineno is None or not node.lineno <= line_number <= node.end_lineno:
                    continue
                block = ast.get_source_segment(source, node)
                if block:
                    block_lines: list[str] = []
                    for n, block_line in enumerate(block.splitlines()):
                        if n == line_number - node.lineno:
                            block_lines.append(f"> {block_line}")
                        else:
                            block_lines.append(f"  {block_line}")
                    line = "\n".join(block_lines)
                break
        else:
            line = f"> {line.strip()}"
        return line, template

    def _indent(self, indent: int, text: str) -> str:
        return "\n".join(f"{' ' * indent}{line}" for line in text.splitlines())


from .junk import Junk



import contextlib
import pathlib
from typing import Any, Iterator

from auryn import Junk, Lines
from auryn.utils import and_, is_path

UNDEFINED = object()
DEFINITIONS = "definitions"
PARAMETERS = "parameters"
BOOKMARKS = "bookmarks"


class Bookmark:

    def __init__(self, indent: int) -> None:
        self.indent = indent
        self.lines: list[Any] = []

    def __str__(self) -> str:
        return "".join(map(str, self.lines))


def meta_eval(junk: Junk, code: str) -> None:
    code = eval(f"f{code!r}", junk.meta_namespace)
    junk.emit_code(code)
    junk.proceed()
    with junk.increase_code_indent():
        junk.proceed(junk.line.children.snap())


def meta_emit(junk: Junk, text: str) -> None:
    text = eval(f"f{text!r}", junk.meta_namespace)
    junk.emit_text(junk.line.indent, text)
    junk.proceed()


def meta_include(
    junk: Junk,
    template: str | pathlib.Path,
    load: str | pathlib.Path | dict[str, Any] | None = None,
    render: bool = True,
    interpolate: bool = True,
    with_namespace: bool = False,
) -> None:
    if not render:
        if is_path(template, junk.path.parent):
            path = junk.path.parent / template
            junk.emit_text_block(junk.line.indent, path.read_text(), interpolate=interpolate)
        else:
            junk.emit_text_block(junk.line.indent, str(template), interpolate=interpolate)
        return
    included_junk = junk.derive(template, with_namespace=with_namespace)
    if load:
        included_junk.load(load)
    if junk.has_line:
        included_junk.lines.snap(junk.line.indent)
    included_junk.transpile(junk.meta_context)
    junk.emit_code(included_junk.to_string(), add_source_comment=False)


def meta_define(junk: Junk, name: str) -> None:
    definitions: dict[str, Lines] = junk.meta_state.setdefault(DEFINITIONS, {})
    definitions[name] = junk.line.children


def meta_insert(junk: Junk, name: str, required: bool = False) -> None:
    definitions: dict[str, Lines] = junk.meta_state.get(DEFINITIONS, {})
    if name not in definitions:
        if required:
            raise ValueError(
                f"missing required definition {name!r} on {junk.line} "
                f"(available definitions are {and_(definitions)})"
            )
        junk.proceed(junk.line.children.snap())
        return
    junk.proceed(definitions[name].snap(junk.line.indent))


def meta_extend(junk: Junk, template: str | pathlib.Path) -> None:
    if junk.line.children:
        with junk.patch(meta_lines=junk.meta_lines, code_indent=junk.code_indent):
            junk.proceed(junk.line.children.snap())
        meta_include(junk, template)
        return

    def replace_code(junk: Junk) -> None:
        junk.meta_lines.clear()
        junk.code_indent = 0
        meta_include(junk, template)

    junk.meta_callbacks.append(replace_code)


def meta_interpolate(junk: Junk, interpolation: str) -> None:
    if junk.line.children:
        with junk.patch(interpolation=interpolation):
            junk.proceed(junk.line.children.snap())
    else:
        junk.interpolation = interpolation


def meta_raw(junk: Junk) -> None:
    if junk.line.children:
        text = junk.line.children.snap().to_string()
        junk.emit_text_block(0, text, interpolate=False)
    else:
        junk.transpilers.clear()
        junk.transpilers[""] = emit_raw


def emit_raw(junk: Junk, content: str) -> None:
    junk.emit_code(f"{junk.EMIT}({junk.line.indent}, {content!r})")
    junk.proceed()


def meta_stop(junk: Junk) -> None:
    junk.emit_code(f"raise {junk.STOP_EVALUATION}()")


def meta_param(junk: Junk, name: str, default: Any = UNDEFINED) -> None:
    parameters: dict[str, Any] = junk.meta_state.setdefault(PARAMETERS, {})
    parameters[name] = default if default is not UNDEFINED else "<required>"
    if default is UNDEFINED:
        message = f"missing required parameter {name!r} in {junk.source}"
        junk.emit_code(
            f"""
            if {name!r} not in globals():
                raise ValueError({message!r})
            """
        )
    else:
        junk.emit_code(
            f"""
            try:
                {name}
            except NameError:
                {name} = {default!r}
            """
        )


def meta_inline(junk: Junk) -> None:
    junk.emit_text(junk.line.indent, "", newline=False)
    with junk.patch(inline=True):
        junk.proceed()
    with junk.patch(inline=False):
        junk.emit_text(None, "")


def meta_assign(junk: Junk, name: str) -> None:
    junk.emit_code("with assign() as _:")
    with junk.increase_code_indent():
        junk.proceed(junk.line.children.snap())
    junk.emit_code(f"{name} = ''.join(_).strip()")


@contextlib.contextmanager
def eval_assign(junk: Junk) -> Iterator[list[str]]:
    eval_lines: list[str] = []
    with junk.patch(eval_lines=eval_lines):
        yield eval_lines


def meta_bookmark(junk: Junk, name: str) -> None:
    junk.emit_code(f"bookmark({junk.line.indent}, {junk.interpolate(name)})")


def eval_bookmark(junk: Junk, indent: int, name: str) -> None:
    bookmark = Bookmark(indent)
    bookmarks: dict[str, Bookmark] = junk.meta_state.setdefault(BOOKMARKS, {})
    bookmarks[name] = bookmark
    junk.eval_lines.append(bookmark)


def meta_append(junk: Junk, name: str) -> None:
    junk.emit_code(f"with append({junk.interpolate(name)}, {str(junk.line)!r}):")
    with junk.increase_code_indent():
        junk.proceed(junk.line.children.snap(0))


@contextlib.contextmanager
def eval_append(junk: Junk, name: str, line: str) -> Iterator[None]:
    bookmarks: dict[str, Bookmark] = junk.meta_state.get(BOOKMARKS, {})
    if name not in bookmarks:
        raise ValueError(
            f"missing bookmark {name!r} referenced on {line} " f"(available bookmarks are {and_(bookmarks)})"
        )
    bookmark = bookmarks[name]
    with junk.patch(eval_lines=bookmark.lines, eval_indent=bookmark.indent):
        yield


def meta_strip(junk: Junk, suffix: str) -> None:
    junk.emit_code(f"strip({suffix!r})")


def eval_strip(junk: Junk, suffix: str) -> None:
    junk.eval_lines[-1] = junk.eval_lines[-1].rstrip().strip(suffix)


def eval_camel_case(junk: Junk, name: str) -> str:
    return "".join(word.capitalize() for word in name.split("_"))


def eval_concat(junk: Junk, *args: Any) -> str:
    return "".join(map(str, args))



import pathlib
from typing import Any

from .junk import Junk


def transpile(
    template: str | pathlib.Path,
    meta_context: dict[str, Any] | None = None,
    /,
    *,
    load_common: bool | None = None,
    add_source_comments: bool | None = None,
    load: str | pathlib.Path | dict[str, Any] | None = None,
    standalone: bool | None = None,
    **meta_context_kwargs: Any,
) -> str:
    junk = Junk(template, load_common=load_common, add_source_comments=add_source_comments, stack_level=1)
    if load:
        junk.load(load)
    junk.transpile(meta_context, **meta_context_kwargs)
    return junk.to_string(standalone=standalone)


def render(
    template: str | pathlib.Path,
    context: dict[str, Any] | None = None,
    /,
    *,
    load_common: bool | None = None,
    load: str | pathlib.Path | dict[str, Any] | None = None,
    meta_context: dict[str, Any] | None = None,
    **context_kwargs: Any,
) -> str:
    junk = Junk(template, load_common=load_common, stack_level=1)
    if load:
        junk.load(load)
    junk.transpile(meta_context)
    return junk.evaluate(context, **context_kwargs)


def evaluate(
    path: str | pathlib.Path,
    context: dict[str, Any] | None = None,
    /,
    **context_kwargs: Any,
) -> str:
    if isinstance(path, str) and "\n" in path:
        meta_lines = path.splitlines()
    else:
        path = pathlib.Path(path)
        meta_lines = path.read_text().splitlines()
    junk = Junk(stack_level=1)
    junk.meta_lines = meta_lines
    return junk.evaluate(context, **context_kwargs)
